/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Tzeggai - additional utility classes
 ******************************************************************************/
package de.schmitzm.jfree.feature.style;

import org.geotools.feature.FeatureCollection;
import org.jfree.chart.JFreeChart;
import org.jfree.data.general.Dataset;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;

import de.schmitzm.jfree.chart.style.AbstractChartStyle;
import de.schmitzm.jfree.chart.style.ChartStyle;
import de.schmitzm.jfree.table.style.TableChartStyle;

/**
 * This interface extends the chart style with several functionalities used to
 * define a chart on a {@link FeatureCollection}.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public interface FeatureChartStyle extends TableChartStyle<FeatureCollection<SimpleFeatureType, SimpleFeature>> {

  /**
   * Creates a chart according to the given feature collection.
   * @param fc a feature collection
   * @exception UnsupportedOperationException if the style can not be applied to
   *              the {@link FeatureCollection} (e.g. the
   *              {@link FeatureCollection} does not provide the required
   *              attributes)
   */
  public JFreeChart applyToFeatureCollection(
      FeatureCollection<SimpleFeatureType, SimpleFeature> fc);

  /**
   * This class defines a dummy implementation of {@link FeatureChartStyle} just
   * to maintain the properties of the interface {@link FeatureChartStyle}, so
   * sub classes of {@link FeatureChartStyle} which usually are derived from a
   * normal {@link ChartStyle} implementation must not implement the
   * {@link FeatureChartStyle} maintenance each. Instead they can create an
   * instance of this dummy an pipe their method implementations to the dummy!<br>
   * <br>
   * The {@link #applyToFeatureCollection(FeatureCollection)} and
   * {@link #applyToDataset(Dataset)} methods are not implemented by the dummy,
   * but throw an exception instead!!
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
   */
  public static class Dummy extends TableChartStyle.Dummy<FeatureCollection<SimpleFeatureType, SimpleFeature>> implements FeatureChartStyle {

    /**
     * Creates a new dummy. This constructor is protected so that only derived
     * classes can instantiate a dummy! There is no limit for the number of
     * attributes which can be defined by the style.
     * @param id a (unique) ID for the style
     */
    protected Dummy(String id) {
      this(id, -1);
    }

    /**
     * Creates a new dummy. This constructor is protected so that only derived
     * classes can instantiate a dummy!
     * @param id a (unique) ID for the style
     * @param maxAttrCount attribute count needed to specify the chart data from
     *          feature collection
     */
    protected Dummy(String id, int maxAttrCount) {
      super(id, maxAttrCount);
    }

    /**
     * Creates a (deep) clone of this style. The properties of the super class (
     * {@link AbstractChartStyle}) are <b>ignored</b> because they are unused
     * for the dummy.
     */
    @Override
    public AbstractChartStyle copy() {
      return (AbstractChartStyle) copyTo((FeatureChartStyle) new FeatureChartStyle.Dummy(""));
    }

    /**
     * Copies all properties of this style to another one. The properties of the
     * super class ({@link AbstractChartStyle}) are <b>ignored</b> because they
     * are unused for the dummy.
     * @param dest destination object (if {@code null} the copy is created by
     *          {@link #copy()})
     * @return {@code dest} or the new instance
     */
    @Override
    public ChartStyle copyTo(ChartStyle dest) {
      // !! do NOT copy the super class properties !!
      // !! copy only this classes properties !!
      // !! Reason: the dummy does not maintain the !!
      // !! super class properties and we !!
      // !! do not want to overwrite them !!
      if (dest instanceof FeatureChartStyle) {
        // No own properties yet for FeatureChartStyle to copy
        // Note: this method must be overwritten anyway, so that the
        //       super-method is NOT called!!
      }

      return dest;
    }


    /**
     * Does nothing, but always throws a {@link UnsupportedOperationException},
     * because the dummy can not provide this functionality.
     */
    @Override
    public JFreeChart applyToFeatureCollection(
        FeatureCollection<SimpleFeatureType, SimpleFeature> fc) {
      throw new UnsupportedOperationException(
          "FeatureChartStyle.Dummy does not implement applyToFeatureCollection(..)");
    }

  }
}
