/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Tzeggai - additional utility classes
 ******************************************************************************/
package de.schmitzm.jfree.feature.style;

import org.geotools.feature.FeatureCollection;
import org.jfree.chart.JFreeChart;
import org.jfree.data.general.Dataset;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;

import de.schmitzm.jfree.chart.style.ChartStyle;
import de.schmitzm.jfree.chart.style.ScatterChartStyle;
import de.schmitzm.jfree.table.style.TableScatterChartStyle;

/**
 * This class extends the {@link ScatterChartStyle} with the properties
 * of the {@link FeatureChartStyle}.<br>
 * After instantiation the 2 feature attributes used to define the chart
 * data must be set by {@link #setAttributeName(int, String)}. Without
 * setting the attributes the style can not be applied to a feature collection!
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 */
public class FeatureScatterChartStyle extends TableScatterChartStyle<FeatureCollection<SimpleFeatureType, SimpleFeature>> implements FeatureChartStyle {
  /** Used to maintain the {@link FeatureChartStyle} properties. */
  protected FeatureChartStyle.Dummy dummyFeatureChartStyle = null;
  
  /**
   * Creates a scatter chart style a regression line shown.
   * @param id   a (unique) ID for the style
   */
  public FeatureScatterChartStyle(String id) {
    super(id);
    dummyFeatureChartStyle = new FeatureChartStyle.Dummy(id,2);
  }
  

  /**
   * Creates a scatter chart style a regression line shown.
   * @param id         a (unique) ID for the style
   * @param title      title for the chart
   * @param xAttrName  feature attribute used for domain axis
   * @param yAttrName  feature attribute used for range axis
   */
  public FeatureScatterChartStyle(String id, String title, String xAttrName, String yAttrName) {
    this(id);
    setAttributeName(0, xAttrName);
    setAttributeName(1, yAttrName);
    getTitleStyle().setLabel(title);
  }

  /**
   * Creates a (deep) clone of this style.
   */
  @Override
  public FeatureScatterChartStyle copy() {
    return (FeatureScatterChartStyle) copyTo( new FeatureScatterChartStyle("") );
  }
  
  /**
   * Copies all properties of this style to another one.
   * @param dest destination object (if {@code null} the copy
   *             is created by {@link #copy()})
   * @return {@code dest} or the new instance
   */
  @Override
  public ChartStyle copyTo(ChartStyle dest) {
    // copy the super class properties
    dest = super.copyTo(dest);
    // copy this classes properties
    if ( dest instanceof FeatureChartStyle )
      dummyFeatureChartStyle.copyTo(dest);

    return dest;
  }

  /**
   * Creates an appropriate {@link Dataset} for the attributes defined
   * by this style.
   * @see #applyToFeatureCollection(FeatureCollection)
   */
  @Override
  public JFreeChart applyToTable(FeatureCollection<SimpleFeatureType, SimpleFeature> fc) {
    return applyToFeatureCollection(fc);
  }


  /**
   * Creates an appropriate {@link Dataset} for the attributes defined
   * by this style (according to the attributes types in the given
   * {@link FeatureCollection}) and calls {@link #applyToDataset(Dataset)}.
   */
  @Override
  public JFreeChart applyToFeatureCollection(FeatureCollection<SimpleFeatureType, SimpleFeature> fc) {
    Dataset[] dataset = FeatureChartUtil.createDataset(fc,this);
    return applyToDataset(dataset);
  }
}
